@@ -1,14 +1,21 @@
 # -*- coding: utf-8 -*-
+
+'''
+http://www.python.org/topics/database/DatabaseAPI-2.0.html
+'''
+
 import datetime
 import glob
 import os
 import sys
 import time
+import traceback
 import warnings
 import decimal
 import binascii
 import jpype
 from .sqlcliexception import SQLCliException
+from .sqloption import SQLOptions
 
 
 def reraise(tp, value, tb=None):
@@ -23,11 +30,15 @@ def reraise(tp, value, tb=None):
 
 string_type = str
 
+
 # Mapping from java.sql.Types attribute name to attribute value
 _jdbc_name_to_const = None
 
+
 # Mapping from java.sql.Types attribute constant value to it's attribute name
-_jdbc_const_to_name = None
+_jdbc_const_to_name = []
+
+_sqloptions = SQLOptions()
 
 
 def _handle_sql_exception_jpype():
@@ -74,7 +85,10 @@ def _jdbc_connect_jpype(jclassname, url, driver_args, jars, libs):
         _init_types(types_map)
 
     # register driver for DriverManager
-    jpype.JClass(jclassname)
+    try:
+        jpype.JClass(jclassname)
+    except TypeError as te:
+        raise SQLCliException("SQLCLI-00000: Load java class failed. [" + str(jclassname) + "]")
     if isinstance(driver_args, dict):
         Properties = jpype.java.util.Properties
         info = Properties()
@@ -83,7 +97,14 @@ def _jdbc_connect_jpype(jclassname, url, driver_args, jars, libs):
         dargs = [info]
     else:
         dargs = driver_args
-    return jpype.java.sql.DriverManager.getConnection(url, *dargs)
+
+    try:
+        return jpype.java.sql.DriverManager.getConnection(url, *dargs)
+    except jpype.java.sql.SQLException as je:
+        raise SQLCliException(je.toString().
+                              replace("java.sql.SQLException: ", "").
+                              replace("java.sql.SQLTransientConnectionException: ", "").
+                              replace("java.sql.SQLInvalidAuthorizationSpecException: ", ""))
 
 
 def _get_classpath():
@@ -150,8 +171,9 @@ class DBAPITypeObject(object):
         try:
             type_name = _jdbc_const_to_name[jdbc_type_const]
         except KeyError:
-            warnings.warn("Unknown JDBC type with constant value %d. "
-                          "Using None as a default type_code." % jdbc_type_const)
+            if "SQLCLI_DEBUG" in os.environ:
+                warnings.warn("Unknown JDBC type with constant value %d. "
+                              "Using None as a default type_code." % jdbc_type_const)
             return None
         try:
             return cls._mappings[type_name]
@@ -235,7 +257,7 @@ Timestamp = _str_func(datetime.datetime)
 
 
 # DB-API 2.0 Module Interface connect constructor
-def connect(jclassname, url, driver_args=None, jars=None, libs=None):
+def connect(jclassname, url, driver_args=None, jars=None, libs=None, sqloptions=None):
     """Open a connection to a database using a JDBC driver and return
     a Connection instance.
 
@@ -266,7 +288,30 @@ def connect(jclassname, url, driver_args=None, jars=None, libs=None):
             libs = [libs]
     else:
         libs = []
-    jconn = _jdbc_connect_jpype(jclassname, url, driver_args, jars, libs)
+
+    # 程序用到的各种会话控制参数
+    global _sqloptions
+    _sqloptions = sqloptions
+
+    # 如果无法连接，则尝试3次，间隔等待2秒
+    jconn = None
+    retryCount = 0
+    while True:
+        try:
+            jconn = _jdbc_connect_jpype(jclassname, url, driver_args, jars, libs)
+            break
+        except SQLCliException as je:
+            if jconn is None:
+                # jconn 为空，可能是网络错误，这里重复尝试
+                if "SQLCLI_DEBUG" in os.environ:
+                    print('traceback.print_exc():\n%s' % traceback.print_exc())
+                    print('traceback.format_exc():\n%s' % traceback.format_exc())
+                retryCount = retryCount + 1
+                if retryCount >= int(_sqloptions.get("CONN_RETRY_TIMES")):
+                    raise je
+                else:
+                    time.sleep(2)
+                    continue
     return Connection(jconn, _converters)
 
 
@@ -303,7 +348,8 @@ class Connection(object):
     def close(self):
         if self._closed:
             raise Error()
-        self.jconn.close()
+        if self.jconn is not None:
+            self.jconn.close()
         self._closed = True
 
     def commit(self):
@@ -388,7 +434,6 @@ class Cursor(object):
 
     def _set_stmt_parms(self, prep_stmt, parameters):
         for i in range(len(parameters)):
-            # print (i, parameters[i], type(parameters[i]))
             prep_stmt.setObject(i + 1, parameters[i])
 
     def execute(self, operation, parameters=None):
@@ -435,13 +480,23 @@ class Cursor(object):
         row = []
         for col in range(1, self._meta.getColumnCount() + 1):
             sqltype = self._meta.getColumnType(col)
-            converter = self._converters.get(sqltype, _unknownSqlTypeConverter)
-            if str(converter.__name__) == "_unknownSqlTypeConverter":
-                if "SQLCLI_DEBUG" in os.environ:
-                    warnings.warn("Unknown JDBC convert with constant value " + str(sqltype) )
+            m_ColumnClassName = self._meta.getColumnClassName(col)
+            if sqltype in self._converters.keys():
+                converter = self._converters.get(sqltype)
+            else:
+                if m_ColumnClassName.upper().find("BFILE") != -1:
+                    converter = _DEFAULT_CONVERTERS["BFILE"]
+                elif m_ColumnClassName in ('oracle.sql.TIMESTAMPTZ', 'oracle.sql.TIMESTAMPLTZ'):
+                    converter = _DEFAULT_CONVERTERS["TIMESTAMP_WITH_TIMEZONE"]
+                else:
+                    converter = _unknownSqlTypeConverter
+                    if "SQLCLI_DEBUG" in os.environ:
+                        warnings.warn("Unknown JDBC convert with constant value " + str(sqltype) +
+                                      ":" + self._meta.getColumnClassName(col))
             if "SQLCLI_DEBUG" in os.environ:
-                print("JDBC SQLType=[" + str(converter.__name__) + "] for col [" + str(col) + "]. sqltype=[" + str(sqltype) + "]")
-            v = converter(self._rs, col)
+                print("JDBC SQLType=[" + str(converter.__name__) + "] for col [" + str(col) + "]. " +
+                      "sqltype=[" + str(sqltype) + ":" + m_ColumnClassName + "]")
+            v = converter(self._connection.jconn, self._rs, col)
             row.append(v)
         return tuple(row)
 
@@ -491,14 +546,14 @@ class Cursor(object):
         self.close()
 
 
-def _unknownSqlTypeConverter(rs, col):
+def _unknownSqlTypeConverter(conn, rs, col):
     java_val = rs.getObject(col)
     if java_val is None:
         return
     return str(java_val)
 
 
-def _to_datetime(rs, col):
+def _to_datetime(conn, rs, col):
     java_val = rs.getTimestamp(col)
     if not java_val:
         return
@@ -507,14 +562,14 @@ def _to_datetime(rs, col):
     return str(d)
 
 
-def _to_time(rs, col):
+def _to_time(conn, rs, col):
     java_val = rs.getTime(col)
     if not java_val:
         return
     return str(java_val)
 
 
-def _to_date(rs, col):
+def _to_date(conn, rs, col):
     java_val = rs.getDate(col)
     if not java_val:
         return
@@ -526,24 +581,28 @@ def _to_date(rs, col):
     return str(java_val)[:10]
 
 
-def _to_binary(rs, col):
+def _to_binary(conn, rs, col):
     java_val = rs.getObject(col)
     if java_val is None:
         return
     m_TypeName = str(java_val.getClass().getTypeName())
     if m_TypeName == "byte[]":
         return binascii.b2a_hex(java_val)
-    elif m_TypeName in ("java.sql.Blob", "oracle.sql.BLOB"):
-        return binascii.b2a_hex(java_val.getBytes(1, 1024))
-    elif m_TypeName.find("JDBCBlobClient") != -1 :
-        return binascii.b2a_hex(java_val.getBytes(1, 1024))
+    elif m_TypeName.upper().find("BLOB") != -1:
+        m_Bytes = java_val.getBytes(1, 1024)
+        if m_Bytes is not None:
+            return binascii.b2a_hex(java_val.getBytes(1, 1024))
+        else:
+            return ""
+    elif m_TypeName.find("BFILE") != -1:
+        return "bfilename(" + java_val.getDirAlias() + ":" + java_val.getName() + ")"
     else:
         # return binascii.b2a_hex(java_val.getBytes(0, 1024))
         raise SQLCliException("SQLCLI-00000: Unknown java class type [" + m_TypeName + "] in _to_binary")
 
 
 def _java_to_py(java_method):
-    def to_py(rs, col):
+    def to_py(conn, rs, col):
         java_val = rs.getObject(col)
         if java_val is None:
             return
@@ -554,47 +613,62 @@ def _java_to_py(java_method):
     return to_py
 
 
-def _java_to_py_bigdecimal():
-    def to_py(rs, col):
-        java_val = rs.getObject(col)
-        if java_val is None:
-            return
-        m_TypeName = str(java_val.getClass().getTypeName())
-        if m_TypeName == "java.math.BigDecimal":
-            return decimal.Decimal(java_val.stripTrailingZeros().toPlainString())
-        elif m_TypeName == "java.lang.Long":
-            return decimal.Decimal(java_val.toString())
-        elif m_TypeName == "java.math.BigInteger":
-            return decimal.Decimal(java_val.toString())
-        else:
-            raise SQLCliException(
-                "SQLCLI-00000: Unknown java class type [" + m_TypeName + "] in _java_to_py_bigdecimal")
-    return to_py
+def _java_to_py_bigdecimal(conn, rs, col):
+    java_val = rs.getObject(col)
+    if java_val is None:
+        return
+    m_TypeName = str(java_val.getClass().getTypeName())
+    if m_TypeName == "java.math.BigDecimal":
+        return decimal.Decimal(java_val.toPlainString())
+    elif m_TypeName == "java.lang.Long":
+        return decimal.Decimal(java_val.toString())
+    elif m_TypeName == "java.math.BigInteger":
+        return decimal.Decimal(java_val.toString())
+    else:
+        raise SQLCliException(
+            "SQLCLI-00000: Unknown java class type [" + m_TypeName + "] in _java_to_py_bigdecimal")
 
 
-def _java_to_py_timestampwithtimezone():
-    def to_py(rs, col):
-        java_val = rs.getObject(col)
-        if java_val is None:
-            return
-        m_TypeName = str(java_val.getClass().getTypeName())
-        if m_TypeName == "java.time.OffsetDateTime":
-            return java_val.format(
-                jpype.java.time.format.DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm:ss.SSSSSS Z"))
-        else:
-            raise SQLCliException(
-                "SQLCLI-00000: Unknown java class type [" + m_TypeName +
-                "] in _java_to_py_timestampwithtimezone")
-    return to_py
+def _java_to_py_timestampwithtimezone(conn, rs, col):
+    java_val = rs.getObject(col)
+    if java_val is None:
+        return
+    m_TypeName = str(java_val.getClass().getTypeName())
+    if m_TypeName == "java.time.OffsetDateTime":
+        return java_val.format(
+            jpype.java.time.format.DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm:ss.SSSSSS Z"))
+    elif m_TypeName in ('oracle.sql.TIMESTAMPTZ', 'oracle.sql.TIMESTAMPLTZ'):
+        return java_val.offsetDateTimeValue(conn).format(
+            jpype.java.time.format.DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm:ss.SSSSSS Z"))
+    else:
+        raise SQLCliException(
+            "SQLCLI-00000: Unknown java class type [" + m_TypeName +
+            "] in _java_to_py_timestampwithtimezone")
 
 
-def _java_to_py_str():
-    def to_py(rs, col):
-        java_val = rs.getObject(col)
-        if java_val is None:
-            return
-        return str(java_val)
-    return to_py
+def _java_to_py_clob(conn, rs, col):
+    java_val = rs.getObject(col)
+    if java_val is None:
+        return
+    m_TypeName = str(java_val.getClass().getTypeName())
+    if m_TypeName.upper().find('CLOB') != -1:
+        m_Length = java_val.length()
+        m_TrimLength = int(_sqloptions.get("LOB_LENGTH"))
+        m_ColumnValue = java_val.getSubString(1, m_TrimLength)
+        if m_Length > int(m_TrimLength):
+            m_ColumnValue = m_ColumnValue + "..."
+        return m_ColumnValue
+    else:
+        raise SQLCliException(
+            "SQLCLI-00000: Unknown java class type [" + m_TypeName +
+            "] in _java_to_py_clob")
+
+
+def _java_to_py_str(conn, rs, col):
+    java_val = rs.getObject(col)
+    if java_val is None:
+        return
+    return str(java_val)
 
 
 def _init_types(types_map):
@@ -606,15 +680,13 @@ def _init_types(types_map):
 
 
 def _init_converters(types_map):
-    """Prepares the converters for conversion of java types to python
-    objects.
-    types_map: Mapping of java.sql.Types field name to java.sql.Types
-    field constant value"""
+    # 存在不属于标准JDBC的convert情况，所以如果types_map找不到，这里不报错
     global _converters
     _converters = {}
     for i in _DEFAULT_CONVERTERS:
-        const_val = types_map[i]
-        _converters[const_val] = _DEFAULT_CONVERTERS[i]
+        if i in types_map:
+            const_val = types_map[i]
+            _converters[const_val] = _DEFAULT_CONVERTERS[i]
 
 
 # Mapping from java.sql.Types field to converter method
@@ -624,10 +696,12 @@ _DEFAULT_CONVERTERS = {
     # see
     # http://download.oracle.com/javase/8/docs/api/java/sql/Types.html
     # for possible keys
-    'CHAR': _java_to_py_str(),
-    'LONGVARCHAR': _java_to_py_str(),
-    'VARCHAR':  _java_to_py_str(),
-    'TIMESTAMP_WITH_TIMEZONE': _java_to_py_timestampwithtimezone(),
+    'CHAR': _java_to_py_str,
+    'LONGVARCHAR': _java_to_py_str,
+    'VARCHAR':  _java_to_py_str,
+    'NCLOB':  _java_to_py_clob,
+    'CLOB':  _java_to_py_clob,
+    'TIMESTAMP_WITH_TIMEZONE': _java_to_py_timestampwithtimezone,
     'TIMESTAMP': _to_datetime,
     'TIME': _to_time,
     'DATE': _to_date,
@@ -635,15 +709,16 @@ _DEFAULT_CONVERTERS = {
     'BINARY': _to_binary,
     'LONGVARBINARY': _to_binary,
     'BLOB': _to_binary,
-    'DECIMAL': _java_to_py_bigdecimal(),
-    'NUMERIC': _java_to_py_bigdecimal(),
+    'BFILE': _to_binary,
+    'DECIMAL': _java_to_py_bigdecimal,
+    'NUMERIC': _java_to_py_bigdecimal,
     'DOUBLE': _java_to_py('doubleValue'),
     'FLOAT': _java_to_py('doubleValue'),
     'REAL': _java_to_py('doubleValue'),
     'TINYINT': _java_to_py('intValue'),
     'INTEGER': _java_to_py('intValue'),
     'SMALLINT': _java_to_py('intValue'),
-    'BIGINT':  _java_to_py_bigdecimal(),
+    'BIGINT':  _java_to_py_bigdecimal,
     'BOOLEAN': _java_to_py('booleanValue'),
     'BIT': _java_to_py('booleanValue')
 }
